package cz.upol.pja.lecture10.jbc;

import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.bytecode.BadBytecode;
import javassist.bytecode.Bytecode;
import javassist.bytecode.ConstPool;
import javassist.bytecode.MethodInfo;

/**
 * Trida poskytujici rozhrani pro definici jednotlivych metod
 */
public class MethodBuilder {

	private final CtClass generatedClass;
	private final CtBehavior method;
	private final MethodInfo methodInfo;
	private Bytecode bytecode;
	private final String methodName;

	// konstruktor by mel byt volan pouze z tridy ClassBuilder 
	MethodBuilder(CtClass generatedClass, String methodName, int modifiers, CtClass returnType, CtClass[] paramTypes) {
		this.generatedClass = generatedClass;
		
		if (methodName != null) {
			this.method = new CtMethod(returnType, methodName, paramTypes, generatedClass);
		} else {
			this.method = new CtConstructor(paramTypes, generatedClass);
		}
		this.method.setModifiers(modifiers);
		this.methodInfo = this.method.getMethodInfo();
		this.methodName = methodName;
	}
	
	/**
	 * Vraci objekt, kterym se generuji jednotlive operace v bytecode
	 * @param stackSize -- velikost zasobniku operandu
	 * @param locals -- pocet argumentu + lokalnich promennych
	 * (u nestatickych metod se this bere jako prvni argument) 
	 */
	public Bytecode createByteCode(int stackSize, int locals) {
		if (bytecode != null) throw new RuntimeException("Bytecode already exists");
		bytecode = new Bytecode(methodInfo.getConstPool(), stackSize, locals);
		return bytecode;
	}
	
	/** ulozi zmeny v bytecode k prislusne metode */
	public void save() throws BadBytecode  {
		methodInfo.setCodeAttribute(bytecode.toCodeAttribute());
		methodInfo.rebuildStackMap(generatedClass.getClassPool());
	}
	
	/** vraci odkaz na pool konstant */
	public ConstPool getConstantPool() {
		return methodInfo.getConstPool();
	}

	/** vraci odkaz na tridu, kam metoda patri */
	public CtClass getGeneratedClass() {
		return generatedClass;
	}

	/** vraci true, pokud se jedna o metodu, v pripade konstruktoru vraci false */
	public boolean isMethod() {
		return methodName != null;
	}

	CtBehavior getMethod() {
		return method;
	}
}
